package inf101.v18.rogue101.map;

import inf101.v18.gfx.gfxmode.ITurtle;
import inf101.v18.gfx.textmode.Printer;
import inf101.v18.grid.GridDirection;
import inf101.v18.grid.IArea;
import inf101.v18.grid.IMultiGrid;
import inf101.v18.grid.Location;
import inf101.v18.grid.MultiGrid;
import inf101.v18.rogue101.Main;
import inf101.v18.rogue101.examples.Carrot;
import inf101.v18.rogue101.game.IllegalMoveException;
import inf101.v18.rogue101.items.buff.Shield;
import inf101.v18.rogue101.items.consumable.Manga;
import inf101.v18.rogue101.items.container.Backpack;
import inf101.v18.rogue101.items.container.Chest;
import inf101.v18.rogue101.items.weapon.BasicBow;
import inf101.v18.rogue101.items.weapon.BasicSword;
import inf101.v18.rogue101.items.weapon.FireStaff;
import inf101.v18.rogue101.object.Actor;
import inf101.v18.rogue101.object.FakeWall;
import inf101.v18.rogue101.object.Item;
import inf101.v18.rogue101.object.PlayerCharacter;
import inf101.v18.rogue101.object.Wall;
import javafx.scene.canvas.GraphicsContext;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashSet;
import java.util.IdentityHashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Queue;
import java.util.Random;
import java.util.Set;

public class AGameMap implements GameMap {
    /**
     * The grid that makes up our map
     */
    private final IMultiGrid<Item> grid;
    /**
     * These locations have changed, and need to be redrawn
     */
    private final Set<Location> dirtyLocs = new HashSet<>();
    /**
     * An index of all the items in the map and their locations.
     */
    // an IdentityHashMap uses object identity as a lookup key, so items are
    // considered equal if they are the same object (a == b)
    private final Map<Item, Location> items = new IdentityHashMap<>();

    public AGameMap(IArea area) {
        grid = new MultiGrid<>(area);
    }

    public AGameMap(int width, int height) {
        grid = new MultiGrid<>(width, height);
    }

    @Override
    public void add(Location loc, Item item) {

        // keep track of location of all items
        items.put(item, loc);
        // also keep track of whether we need to redraw this cell
        dirtyLocs.add(loc);

        // do the actual adding
        List<Item> list = grid.get(loc);

        for (int i = 0; i < list.size(); i++) {
            if (item.compareTo(list.get(i)) >= 0) {
                list.add(i, item);
                return;
            }
        }
        list.add(item);
    }

    public void addRandomItems(Location loc) {
        if (!this.hasActors(loc) && !this.hasWall(loc)) {
            Random random = new Random();
            if (random.nextInt(10) < 2) {
                this.add(loc, new Carrot());
            }
            if (random.nextInt(10) < 1) {
                this.add(loc, new Manga());
            }
            if (random.nextInt(2) < 1) {
                this.add(loc, new Chest());
            }
            if (random.nextInt(2) < 1) {
                this.add(loc, new BasicSword());
            }
            if (random.nextInt(2) < 1) {
                this.add(loc, new Shield());
            }
            if (random.nextInt(2) < 1) {
                this.add(loc, new FireStaff());
            }
            if (random.nextInt(2) < 1) {
                this.add(loc, new BasicBow());
            }
            if (random.nextInt(2) < 1) {
                this.add(loc, new Backpack());
            }
        }
    }

    @Override
    public boolean canGo(Location to) {
        return !grid.contains(to, (i) -> ((i instanceof Wall && !(i instanceof FakeWall)) || i instanceof Actor));
    }

    @Override
    public boolean hasNeighbour(Location from, GridDirection dir) {
        return from.canGo(dir);
    }

    @Override
    public boolean canGo(Location from, GridDirection dir) {
        if (!from.canGo(dir)) {
            return false;
        }
        Location loc = from.go(dir);
        return canGo(loc);
    }

    @Override
    public void draw(ITurtle painter, Printer printer) {
        Iterable<Location> cells;
        if (Main.MAP_DRAW_ONLY_DIRTY_CELLS) {
            if (dirtyLocs.isEmpty()) {
                return;
            } else {
                cells = dirtyLocs;
            }
        } else {
            cells = grid.locations();
            painter.as(GraphicsContext.class).clearRect(0, 0, getWidth() * printer.getCharWidth(),
                    getHeight() * printer.getCharHeight());
            printer.clearRegion(1, 1, getWidth(), getHeight());
        }
        GraphicsContext ctx = painter.as(GraphicsContext.class);
        double h = printer.getCharHeight();
        double w = printer.getCharWidth();
        if (Main.MAP_AUTO_SCALE_ITEM_DRAW) {
            ctx.save();
            ctx.scale(w / h, 1.0);
            w = h;
        }
        try {
            for (Location loc : cells) {
                List<Item> list = grid.get(loc);
                String sym = " ";
                if (!list.isEmpty()) {
                    if (Main.MAP_DRAW_ONLY_DIRTY_CELLS) {
                        ctx.clearRect(loc.getX() * w, loc.getY() * h, w, h);
                        // ctx.fillRect(loc.getX() * w, loc.getY() * h, w, h);
                    }
                    painter.save();
                    painter.jumpTo((loc.getX() + 0.5) * w, (loc.getY() + 0.5) * h);
                    boolean dontPrint = list.get(0).draw(painter, w, h);
                    painter.restore();
                    if (!dontPrint) {
                        sym = list.get(0).getPrintSymbol();
                    }
                }
                printer.printAt(loc.getX() + 1, loc.getY() + 1, sym);
            }
        } finally {
            if (Main.MAP_AUTO_SCALE_ITEM_DRAW) {
                ctx.restore();
            }
        }
        dirtyLocs.clear();
    }

    /**
     * Writes a black void unto the whole map.
     * Properly draws only the tiles visible to the player.
     *
     * @param painter A painter
     * @param printer A printer
     */
    public void drawVisible(ITurtle painter, Printer printer) {
        Iterable<Location> cells;
        if (Main.MAP_DRAW_ONLY_DIRTY_CELLS) {
            if (dirtyLocs.isEmpty()) {
                return;
            } else {
                cells = dirtyLocs;
            }
        } else {
            cells = grid.locations();
            painter.as(GraphicsContext.class).clearRect(0, 0, getWidth() * printer.getCharWidth(),
                    getHeight() * printer.getCharHeight());
            printer.clearRegion(1, 1, getWidth(), getHeight());
        }
        GraphicsContext ctx = painter.as(GraphicsContext.class);
        double h = printer.getCharHeight();
        double w = printer.getCharWidth();
        if (Main.MAP_AUTO_SCALE_ITEM_DRAW) {
            ctx.save();
            ctx.scale(w / h, 1.0);
            w = h;
        }
        try {
            PlayerCharacter player = null;
            Location playerPos = null;
            for (Location loc : cells) {
                printer.printAt(loc.getX() + 1, loc.getY() + 1, " ");
                //We need to get the player and its location from somewhere.
                if (this.hasActors(loc) && this.getActors(loc).get(0) instanceof PlayerCharacter) {
                    player = (PlayerCharacter) this.getActors(loc).get(0);
                    playerPos = loc;
                }
            }
            if (player == null) {
                return;
            }
            List<Location> positions = getVisible(getNeighbourhood(playerPos, player.getVision()), playerPos);
            positions.add(playerPos);
            for (Location loc : positions) {
                List<Item> list = grid.get(loc);
                String sym = ".";
                if (!list.isEmpty()) {
                    if (Main.MAP_DRAW_ONLY_DIRTY_CELLS) {
                        ctx.clearRect(loc.getX() * w, loc.getY() * h, w, h);
                        // ctx.fillRect(loc.getX() * w, loc.getY() * h, w, h);
                    }
                    painter.save();
                    painter.jumpTo((loc.getX() + 0.5) * w, (loc.getY() + 0.5) * h);
                    boolean dontPrint = list.get(0).draw(painter, w, h);
                    painter.restore();
                    if (!dontPrint) {
                        sym = list.get(0).getPrintSymbol();
                    }

                }
                printer.printAt(loc.getX() + 1, loc.getY() + 1, sym);
            }
        } finally {
            if (Main.MAP_AUTO_SCALE_ITEM_DRAW) {
                ctx.restore();
            }
        }
        dirtyLocs.clear();
    }

    private List<Location> getVisible(List<Location> neighbours, Location loc) {
        List<Location> invalid = new ArrayList<>();
        for (Location neighbour : neighbours) {
            if (!hasWall(neighbour)) {
                for (Location tile : loc.gridLineTo(neighbour)) {
                    if (hasWall(tile)) {
                        invalid.add(neighbour);
                        break;
                    }
                }
            }
        }
        neighbours.removeAll(invalid);
        return neighbours;
    }

    @Override
    public List<Actor> getActors(Location loc) {
        List<Actor> items = new ArrayList<>();
        for (Item item : grid.get(loc)) {
            if (item instanceof Actor) {
                items.add((Actor) item);
            }
        }

        return items;
    }

    @Override
    public List<Item> getAll(Location loc) {
        return Collections.unmodifiableList(grid.get(loc));
    }

    @Override
    public List<Item> getAllModifiable(Location loc) {
        dirtyLocs.add(loc);
        return grid.get(loc);
    }

    @Override
    public void clean(Location loc) {
        // remove any items that have health < 0:
        if (grid.get(loc).removeIf((item) -> {
            if (item.isDestroyed()) {
                items.remove(item);
                return true;
            } else {
                return false;
            }
        })) {
            dirtyLocs.add(loc);
        }
    }

    @Override
    public IArea getArea() {
        return grid.getArea();
    }

    @Override
    public int getHeight() {
        return grid.getHeight();
    }

    @Override
    public List<Item> getItems(Location loc) {
        List<Item> items = new ArrayList<>(grid.get(loc));
        items.removeIf((i) -> i instanceof Actor);
        return items;
    }

    @Override
    public Location getLocation(Item item) {
        return items.get(item);
    }

    @Override
    public Location getLocation(int x, int y) {
        return grid.getArea().location(x, y);
    }

    @Override
    public Location getNeighbour(Location from, GridDirection dir) {
        if (!hasNeighbour(from, dir)) {
            return null;
        } else {
            return from.go(dir);
        }
    }

    @Override
    public int getWidth() {
        return grid.getWidth();
    }

    @Override
    public Location go(Location from, GridDirection dir) throws IllegalMoveException {
        if (!from.canGo(dir)) {
            throw new IllegalMoveException("Cannot move outside map!");
        }
        Location loc = from.go(dir);
        if (!canGo(loc)) {
            throw new IllegalMoveException("Occupied!");
        }
        return loc;
    }

    @Override
    public boolean has(Location loc, Item target) {
        return grid.contains(loc, target);
    }

    @Override
    public boolean hasActors(Location loc) {
        return grid.contains(loc, (i) -> i instanceof Actor);
    }

    @Override
    public boolean hasItems(Location loc) {
        // true if grid cell contains an item which is not an IActor
        return grid.contains(loc, (i) -> !(i instanceof Actor));
    }

    @Override
    public boolean hasWall(Location loc) {
        return grid.contains(loc, (i) -> i instanceof Wall);
    }

    @Override
    public void remove(Location loc, Item item) {
        grid.remove(loc, item);
        items.remove(item);
        dirtyLocs.add(loc);
    }

    @Override
    public List<Location> getNeighbourhood(Location startLocation, int distance) {
        if (distance < 0 || startLocation == null) {
            throw new IllegalArgumentException();
        } else if (distance == 0) {
            return new ArrayList<>(); // empty!
        }

        // Run BFS to find all neighbors within the specified distance
        Queue<Location> queue = new LinkedList<>();
        List<Location> explored = new ArrayList<>();
        explored.add(startLocation);
        queue.add(startLocation);

        while (!queue.isEmpty()) {
            Location location = queue.poll();

            // The required distance has been reached
            if (Math.abs(startLocation.getX() - location.getX()) >= distance ||
                    Math.abs(startLocation.getY() - location.getY()) >= distance) {
                break;
            }

            for (GridDirection gridDirection : GridDirection.EIGHT_DIRECTIONS) {
                addIfValid(queue, explored, location, gridDirection);
            }
        }

        // While the BFS requires the start location, it's not considered its own neighbor
        explored.remove(startLocation);
        return explored;
    }

    /**
     * Adds the specified location to the queue if valid
     *
     * @param queue         <p>The queue to add the location to</p>
     * @param explored      <p>The already explored locations</p>
     * @param startLocation <p>The location to start at</p>
     * @param direction     <p>The direction to go</p>
     */
    private void addIfValid(Queue<Location> queue, List<Location> explored, Location startLocation,
                            GridDirection direction) {
        Location location;

        // If going in the specified direction would exit the map, just return
        try {
            location = startLocation.go(direction);
        } catch (IndexOutOfBoundsException ignored) {
            return;
        }

        if (!explored.contains(location) && grid.isValid(location)) {
            explored.add(location);
            queue.add(location);
        }
    }

}
